
/*

  []---------------------------------------------------------------------[]
  |                                SIST.C                                 |
  |                                                                       |
  |      Arquivo de inclusao para resolucao de sistemas nao-lineares      |
  |                                                                       |
  |                 (C) 1995, Lenimar Nunes de Andrade                    |
  []---------------------------------------------------------------------[]

*/

/* ----------------------------------------------------------------------- */

int Inversa(float m[3][3]) {

/* Calcula a matriz inversa de m 3 x 3. Devolve 0 se a matriz nao tiver
   inversa ou 1 se a inversa existir. */

  float Det_M, a, b, c, d, e, f, g, h, i;
  const float epsilon = 1E-10;

  a = m[0][0]; b = m[0][1]; c = m[0][2];
  d = m[1][0]; e = m[1][1]; f = m[1][2];
  g = m[2][0]; h = m[2][1]; i = m[2][2];

  Det_M = b*f*g + c*d*h + a*e*i - c*e*g - a*f*h - b*d*i;

  if (fabs(Det_M) <= epsilon) return 0;

  m[0][0] = (e*i - f*h)/Det_M;
  m[0][1] = (c*h - b*i)/Det_M;
  m[0][2] = (b*f - c*e)/Det_M;

  m[1][0] = (f*g - d*i)/Det_M;
  m[1][1] = (a*i - c*g)/Det_M;
  m[1][2] = (c*d - a*f)/Det_M;

  m[2][0] = (d*h - e*g)/Det_M;
  m[2][1] = (b*g - a*h)/Det_M;
  m[2][2] = (a*e - b*d)/Det_M;

  return 1;
}

/* ----------------------------------------------------------------------- */

void MatrizJacobiana(float m[3][3],  float (*f1)(float, float, float),
     float (*f2)(float, float, float), float (*f3)(float, float, float),
							  float x[3]) {

/* Calcula a matriz jacobina J   = (df /dx )
			      ij      i   j 3x3    */

  const float h = 1E-6;

  m[0][0] = (f1(x[0] + h, x[1], x[2]) - f1(x[0] - h, x[1], x[2]))/(2.0*h);
  m[0][1] = (f1(x[0], x[1] + h, x[2]) - f1(x[0], x[1] - h, x[2]))/(2.0*h);
  m[0][2] = (f1(x[0], x[1], x[2] + h) - f1(x[0], x[1], x[2] - h))/(2.0*h);

  m[1][0] = (f2(x[0] + h, x[1], x[2]) - f2(x[0] - h, x[1], x[2]))/(2.0*h);
  m[1][1] = (f2(x[0], x[1] + h, x[2]) - f2(x[0], x[1] - h, x[2]))/(2.0*h);
  m[1][2] = (f2(x[0], x[1], x[2] + h) - f2(x[0], x[1], x[2] - h))/(2.0*h);

  m[2][0] = (f3(x[0] + h, x[1], x[2]) - f3(x[0] - h, x[1], x[2]))/(2.0*h);
  m[2][1] = (f3(x[0], x[1] + h, x[2]) - f3(x[0], x[1] - h, x[2]))/(2.0*h);
  m[2][2] = (f3(x[0], x[1], x[2] + h) - f3(x[0], x[1], x[2] - h))/(2.0*h);
}

/* ----------------------------------------------------------------------- */

/* Calcula prod[] = m[][] . v[] */

#define MATRIZVEZESVETOR(m,v,p)  p[0]=m[0][0]*v[0]+m[0][1]*v[1]+m[0][2]*v[2]; \
				 p[1]=m[1][0]*v[0]+m[1][1]*v[1]+m[1][2]*v[2]; \
				 p[2]=m[2][0]*v[0]+m[2][1]*v[1]+m[2][2]*v[2];

/* ----------------------------------------------------------------------- */

/* Calcula dif[] = v1[] - v2[] */

#define SUBTRAIVETORES(v1, v2, dif)  dif[0]=v1[0]-v2[0]; \
		 dif[1]=v1[1]-v2[1]; dif[2]=v1[2]-v2[2];

/* ----------------------------------------------------------------------- */

/* Calcula o vetor v[] = (f(x[]), g(x[]), h(x[])) */

#define CALCULAVETORF(v, f, g, h, x)   v[0] = f(x[0], x[1], x[2]); \
	  v[1] = g(x[0], x[1], x[2]);  v[2] = h(x[0], x[1], x[2]);

/* ----------------------------------------------------------------------- */

/* Calcula a distancia entre x[] e y[] */

#define QUAD(x) ((x)*(x))

#define DISTANCIA(x0,x1,x2,y0,y1,y2) sqrt(QUAD(x0-y0)+QUAD(x1-y1)+QUAD(x2-y2))

/* ----------------------------------------------------------------------- */

/* Iguala o vetor x[] a y[]; e' o mesmo que memcpy(x, y, 3*sizeof(float)) */

#define IGUALAVETOR(x, y)  x[0] = y[0]; x[1] = y[1]; x[2] = y[2]

/* ----------------------------------------------------------------------- */

int RSist_Newton(float (*f)(float, float, float),
		 float (*g)(float, float, float),
		 float (*h)(float, float, float),
		 float x_inicial[3],
		 float erro,
		 int max_iteracoes,
		 float solucao[3]) {

/* Resolve o sistema de 3 equacoes
			   
			     f(x, y, z) = 0
			     g(x, y, z) = 0
			     h(x, y, z) = 0
			   
  usando o metodo de Newton.

  Se houver erro, e' retornado o valor -1; senao, e' retornado a quantidade
  de iteracoes utilizadas. Se encontrada, a solucao e' retornada no parame-
  tro solucao[].  Sao usados a aproximacao inicial da solucao  x_inicial[],
  e uma quantidade maxima de iteracoes max_iteracoes.

  Ex.: long double x_0[3] = {1.0, 1.0, 1.0}, X[3];
       if (RSist_Newton(F, G, H, x_0, 1E-10, 1000, X) < 0)
	 printf("Sistema nao pode ser resolvido");
       else
	 printf("Solucao encontrada = (%Lf, %Lf, %Lf)", X[0], X[1], X[2]);
*/

  float mat[3][3], prod[3], F[3], dist_vetores = 0;
  int quant_iteracoes = 0;

  do {
    CALCULAVETORF(F, f, g, h, x_inicial);
    MatrizJacobiana(mat, f, g, h, x_inicial);
    quant_iteracoes++;
    if (Inversa(mat)) {
      MATRIZVEZESVETOR(mat, F, prod);
      SUBTRAIVETORES(x_inicial, prod, solucao);
      dist_vetores = DISTANCIA(x_inicial[0], x_inicial[1], x_inicial[2],
			       solucao[0], solucao[1], solucao[2]);
      IGUALAVETOR(x_inicial, solucao);
    }
    else {
      /* ERRO: Matriz jacobiana nao inversivel */
      return -1;
    }
  } while (dist_vetores > erro && quant_iteracoes < max_iteracoes);

  if (quant_iteracoes >= max_iteracoes) return -1;
  return quant_iteracoes;
}

/* ----------------------------------------------------------------------- */

float CalculaMi(float m[3][3], float F[3]) {

/*			 t           t          t
   Calcula mi = <F, m . m . F>/<m . m . F, m . m . F>  */

  float v[3], x;

  v[0] = (m[0][0]*m[0][0]+m[1][0]*m[1][0]+m[2][0]*m[2][0])*F[0] +
	 (m[0][0]*m[0][1]+m[1][0]*m[1][1]+m[2][0]*m[2][1])*F[1] +
	 (m[0][0]*m[0][2]+m[1][0]*m[1][2]+m[2][0]*m[2][2])*F[2];
  v[1] = (m[0][1]*m[0][0]+m[1][1]*m[1][0]+m[2][1]*m[2][0])*F[0] +
	 (m[0][1]*m[0][1]+m[1][1]*m[1][1]+m[2][1]*m[2][1])*F[1] +
	 (m[0][1]*m[0][2]+m[1][1]*m[1][2]+m[2][1]*m[2][2])*F[2];
  v[2] = (m[0][2]*m[0][0]+m[1][2]*m[1][0]+m[2][2]*m[2][0])*F[0] +
	 (m[0][2]*m[0][1]+m[1][2]*m[1][1]+m[2][2]*m[2][1])*F[1] +
	 (m[0][2]*m[0][2]+m[1][2]*m[1][2]+m[2][2]*m[2][2])*F[2];
  x = (v[0]*F[0] + v[1]*F[1] + v[2]*F[2])/(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
  return x;
}

/* ----------------------------------------------------------------------- */

void MatrizJacobianaTransposta(float m[3][3],  float (*f1)(float, float, float),
     float (*f2)(float, float, float), float (*f3)(float, float, float),
							  float x[3]) {

/* Calcula a transposta da matriz jacobina J   = (df /dx )
					    ij      i   j 3x3    */

  const float h = 1E-6;

  m[0][0] = (f1(x[0] + h, x[1], x[2]) - f1(x[0] - h, x[1], x[2]))/(2.0*h);
  m[1][0] = (f1(x[0], x[1] + h, x[2]) - f1(x[0], x[1] - h, x[2]))/(2.0*h);
  m[2][0] = (f1(x[0], x[1], x[2] + h) - f1(x[0], x[1], x[2] - h))/(2.0*h);

  m[0][1] = (f2(x[0] + h, x[1], x[2]) - f2(x[0] - h, x[1], x[2]))/(2.0*h);
  m[1][1] = (f2(x[0], x[1] + h, x[2]) - f2(x[0], x[1] - h, x[2]))/(2.0*h);
  m[2][1] = (f2(x[0], x[1], x[2] + h) - f2(x[0], x[1], x[2] - h))/(2.0*h);

  m[0][2] = (f3(x[0] + h, x[1], x[2]) - f3(x[0] - h, x[1], x[2]))/(2.0*h);
  m[1][2] = (f3(x[0], x[1] + h, x[2]) - f3(x[0], x[1] - h, x[2]))/(2.0*h);
  m[2][2] = (f3(x[0], x[1], x[2] + h) - f3(x[0], x[1], x[2] - h))/(2.0*h);
}

/* ----------------------------------------------------------------------- */

int RSist_Gradiente(float (*f)(float, float, float),
		    float (*g)(float, float, float),
		    float (*h)(float, float, float),
		    float x_inicial[3],
		    float erro,
		    int max_iteracoes,
		    float solucao[3]) {

/* Resolve o sistema de 3 equacoes
			   
			     f(x, y, z) = 0
			     g(x, y, z) = 0
			     h(x, y, z) = 0
			   
  usando o metodo do gradiente.

  Se houver erro, e' retornado o valor -1; senao, e' retornado a quantidade
  de iteracoes utilizadas. Se encontrada, a solucao e' retornada no parame-
  tro solucao[].  Sao usados a aproximacao inicial da solucao  x_inicial[],
  e uma quantidade maxima de iteracoes max_iteracoes.

  Ex.: long double x_0[3] = {1.0, 1.0, 1.0}, X[3];
       if (RSist_Gradiente(F, G, H, x_0, 1E-10, 1000, X) < 0)
	 printf("Sistema nao pode ser resolvido");
       else
	 printf("Solucao encontrada = (%Lf, %Lf, %Lf)", X[0], X[1], X[2]);
*/

  float mi, mat[3][3], prod[3], F[3], dist_vetores = 0;
  int quant_iteracoes = 0;

  do {
    CALCULAVETORF(F, f, g, h, x_inicial);
    MatrizJacobianaTransposta(mat, f, g, h, x_inicial);
    mi = CalculaMi(mat, F);
    quant_iteracoes++;
    MATRIZVEZESVETOR(mat, F, prod);
    prod[0] *= mi;
    prod[1] *= mi;
    prod[2] *= mi;
    SUBTRAIVETORES(x_inicial, prod, solucao);
    dist_vetores = DISTANCIA(x_inicial[0], x_inicial[1], x_inicial[2],
			     solucao[0], solucao[1], solucao[2]);
    IGUALAVETOR(x_inicial, solucao);

  } while (dist_vetores > erro && quant_iteracoes < max_iteracoes);

  if (quant_iteracoes >= max_iteracoes) return -1;
  return quant_iteracoes;
}

/* ----------------------------------------------------------------------- */

int RSist_NewtonModif(float (*f)(float, float, float),
		      float (*g)(float, float, float),
		      float (*h)(float, float, float),
		      float x_inicial[3],
		      float erro,
		      int max_iteracoes,
		      float solucao[3]) {

/* Resolve o sistema de 3 equacoes
			   
			     f(x, y, z) = 0
			     g(x, y, z) = 0
			     h(x, y, z) = 0
			   
  usando o metodo de Newton modificado.

  Se houver erro, e' retornado o valor -1; senao, e' retornado a quantidade
  de iteracoes utilizadas. Se encontrada, a solucao e' retornada no parame-
  tro solucao[].  Sao usados a aproximacao inicial da solucao  x_inicial[],
  e uma quantidade maxima de iteracoes max_iteracoes.

  Ex.: long double x_0[3] = {1.0, 1.0, 1.0}, X[3];
       if (RSist_NewtonModif(F, G, H, x_0, 1E-10, 1000, X) < 0)
	 printf("Sistema nao pode ser resolvido");
       else
	 printf("Solucao encontrada = (%Lf, %Lf, %Lf)", X[0], X[1], X[2]);
*/

  float mat[3][3], prod[3], F[3], dist_vetores, erro_inicial = 0.1;
  int quant_iteracoes = 0;

  do {
    CALCULAVETORF(F, f, g, h, x_inicial);
    MatrizJacobiana(mat, f, g, h, x_inicial);
    quant_iteracoes++;
    if (Inversa(mat)) {
      MATRIZVEZESVETOR(mat, F, prod);
      SUBTRAIVETORES(x_inicial, prod, solucao);
      dist_vetores = DISTANCIA(x_inicial[0], x_inicial[1], x_inicial[2],
			       solucao[0], solucao[1], solucao[2]);
      IGUALAVETOR(x_inicial, solucao);
    }
    else {
      /* ERRO: Matriz jacobiana nao inversivel */
      return -1;
    }
  } while (dist_vetores > erro_inicial && quant_iteracoes < max_iteracoes);

  do {
    CALCULAVETORF(F, f, g, h, x_inicial);
    quant_iteracoes++;
    MATRIZVEZESVETOR(mat, F, prod);
    SUBTRAIVETORES(x_inicial, prod, solucao);
    dist_vetores = DISTANCIA(x_inicial[0], x_inicial[1], x_inicial[2],
			     solucao[0], solucao[1], solucao[2]);
    IGUALAVETOR(x_inicial, solucao);
  } while (dist_vetores > erro && quant_iteracoes < max_iteracoes);

  if (quant_iteracoes >= max_iteracoes) return -1;
  return quant_iteracoes;
}

/* ----------------------------------------------------------------------- */

/*** FIM DE SIST.C ***/
